/*----------------------

   GATE - Geant4 Application for Tomographic Emission 
   OpenGATE Collaboration 
     
   Giovanni Santin <giovanni.santin@cern.ch>
   Daniel Strul <daniel.strul@iphe.unil.ch> 
   Luc Simon <luc.simon@iphe.unil.ch> 
   Sebastien Jan <sjan@shfj.cea.fr>
     
   Copyright (C) 2002,2003 UNIL/IPHE, CH-1015 Lausanne 
   Copyright (C) 2003 CEA/SHFJ, F-91401 Orsay 

This software is distributed under the terms 
of the GNU Lesser General  Public Licence (LGPL) 
See GATE/LICENSE.txt for further details 
----------------------*/

/*!
  \file GateAnalysis.cc
  
  $Log: GateAnalysis.cc,v $
  Revision 1.1.1.1  2005/02/03 13:09:53  jdbeen
  gate cluster code for G4.6.2

  Revision 1.26  2004/02/13 15:42:08  sjan
  C.Groiselle Modifications about Compton OutPut Variables

  Revision 1.25  2003/06/17 12:00:30  dstrul
  Replaced inclusions of iomanip.h and fstream.h with inclusion of g4std/iomanip and g4std/fstream to suppress compilation warning; Replaced call to std::setw() with call to std::setw() as needed

  Revision 1.24  2003/06/03 12:04:01  sjan
  Compton Counting in the detector

  Revision 1.23  2003/02/25 14:37:47  gsantin
  bug fix for scatter analysis in SPECT

  Revision 1.22  2003/02/17 15:44:40  sjan
  Check the scattering on Compton AND Rayleigh diffusion

  Revision 1.21  2002/12/12 19:03:35  gsantin
  verbose setting forward to the trajectory navigator

  Revision 1.20  2002/11/25 15:16:28  gsantin
  adaptations for SPECT, handling of photon trajectories not only in pairs

  Revision 1.19  2002/10/21 15:23:29  dstrul
  Adaptations for DigiGate: simplified the use of the compilation options, added security for the case when the user tries to run DigiGate without the ROOT option

  Revision 1.18  2002/10/20 13:14:04  dstrul
  GateCrystalHit now provides a method GoodForAnalysis() to tell whether a hit should be analysed then stored into the output hit files

  Revision 1.17  2002/10/20 12:42:30  dstrul
  Adaptations for DigiGate: transportation hits with non-zero energy deposition now are analysed and written to the output hit files

  Revision 1.16  2002/10/19 14:52:10  dstrul
  Adaptations for Digigate: the output modules now have two virtual methods RecordBeginOfAcquisition() and RecordEndOfAcquisition()

  Revision 1.15  2002/10/19 11:49:12  dstrul
  Adaptations for DigiGate: all requests for hit or digi collection are now sent to the output-mgr

  Revision 1.14  2002/10/19 10:50:11  dstrul
  Adaptations for DigiGate: the retrieving of the hits and digis collections is now handled by methods provided by the base-class GateVOutputModule

  Revision 1.13  2002/10/07 09:25:51  gsantin
  NULL volume name for NO-compton case

  Revision 1.12  2002/10/05 22:13:52  gsantin
  verbosity fix

  Revision 1.11  2002/10/04 17:11:28  gsantin
  last compton volume name in the Hits

  Revision 1.10  2002/10/04 13:33:53  lsimon
  Bug fix: GateAnalysis could only detect std-compton events, but not low-energy ones

  Revision 1.9  2002/08/11 12:08:21  dstrul
  Cosmetic cleanup of doxygen documentation

  \brief Class GateAnalysis
  \brief By Giovanni.Santin@cern.ch
  \brief $Id: GateAnalysis.cc,v 1.1.1.1 2005/02/03 13:09:53 jdbeen Exp $
*/

#include "GateAnalysis.hh"
#include "GateAnalysisMessenger.hh"

#include "GateTrajectoryNavigator.hh"

#include "globals.hh"

#include "G4Run.hh"
#include "G4Step.hh"
#include "G4Event.hh"
#include "GateCrystalHit.hh"
#include "GatePhantomHit.hh"
#include "G4VHitsCollection.hh"
#include "G4HCofThisEvent.hh"

#include "G4TrajectoryContainer.hh"

#include "G4VProcess.hh"
#include "RecorderBase.hh"
#include "G4ios.hh"
#include "G4UImanager.hh"
#include "G4RunManager.hh"
#include "GatePrimaryGeneratorAction.hh"

#include "G4ParticleDefinition.hh"
#include "G4Positron.hh"
#include "G4GenericIon.hh"

#include "GateSourceMgr.hh"

#include "G4Navigator.hh"
#include "G4TransportationManager.hh"

#include "GateOutputMgr.hh"

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

GateAnalysis::GateAnalysis(const G4String& name, GateOutputMgr* outputMgr,DigiMode digiMode) 
  : GateVOutputModule(name,outputMgr,digiMode)
{
  m_analysisMessenger = new GateAnalysisMessenger(this);
  m_trajectoryNavigator = new GateTrajectoryNavigator();
  SetVerboseLevel(0);
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo...

GateAnalysis::~GateAnalysis() 
{
  delete m_analysisMessenger;
  if (m_verboseLevel > 0) G4cout << "GateAnalysis deleting..." << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordBeginOfAcquisition()
{
  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordBeginOfAcquisition" << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....


void GateAnalysis::RecordEndOfAcquisition()
{
  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordEndOfAcquisition" << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordBeginOfRun(const G4Run * )
{
  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordBeginOfRun" << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordEndOfRun(const G4Run * )
{
  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordEndOfRun" << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordBeginOfEvent(const G4Event* )
{

//dk
   m_pos_ID=0;
  m_gam1_ID=0;
  m_gam2_ID=0;
//dk_end
      
  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordBeginOfEvent" << G4endl;
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordEndOfEvent(const G4Event* event)
{

  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordEndOfEvent" << G4endl;

  G4TrajectoryContainer* trajectoryContainer = event->GetTrajectoryContainer();

  if (trajectoryContainer)
    m_trajectoryNavigator->SetTrajectoryContainer(trajectoryContainer);
    
  G4int eventID = event->GetEventID();
  G4int runID   = G4RunManager::GetRunManager()->GetCurrentRun()->GetRunID();

  //G4int i;

  if (!trajectoryContainer)
  {
    if (m_verboseLevel > 0) G4cout << "GateAnalysis::RecordEndOfEvent : WARNING : G4TrajectoryContainer not found" << G4endl;
  }
  else
  {

    GateCrystalHitsCollection* CHC = GetOutputMgr()->GetCrystalHitCollection();

    G4int NbHits = 0;
    G4int NpHits = 0;

    if (CHC)
    {    

      NbHits = CHC->entries();
      /*G4cout << "     " << NbHits
	<< " hits are stored in essaiCrystalHitsCollection." << G4endl;*/
    
      //G4int ionID      = 1; // the primary vertex particle
      G4int positronID = 0;
      G4int photon1ID  = 0;
      G4int photon2ID  = 0;
      G4int rootID     = 0;
      G4int primaryID  = 0;

      G4int photon1_phantom_compton = 0;
      G4int photon2_phantom_compton = 0;
        
      G4int photon1_crystal_compton = 0;
      G4int photon2_crystal_compton = 0;

      G4int photon1_phantom_Rayleigh = 0;
      G4int photon2_phantom_Rayleigh = 0;
        
      G4int photon1_crystal_Rayleigh = 0;
      G4int photon2_crystal_Rayleigh = 0;      
 
      // search the positron
      positronID = m_trajectoryNavigator->FindPositronTrackID();

      if (positronID == 0)
      {
	if (m_verboseLevel > 0) G4cout << "GateAnalysis::RecordEndOfEvent : WARNING : positronID == 0" << G4endl;
      }
      if (m_verboseLevel > 1) G4cout 
	<< "GateAnalysis::RecordEndOfEvent : positronID : " << positronID << G4endl;

      //search the two gammas

      std::vector<G4int> photonIDVec = m_trajectoryNavigator->FindAnnihilationGammasTrackID();
      if (photonIDVec.size() == 0)
      {
	// no gamma coming from a positron or an ion, or shooted as primary
	if (m_verboseLevel > 0) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : WARNING : photonIDs not found" << G4endl;
      }
      else 
      {
	if (photonIDVec.size() > 2)
	{
	  G4cout 
	    << "GateAnalysis::RecordEndOfEvent : WARNING : photonID vector size > 2" << G4endl;
	} 
	photon1ID = photonIDVec[0];
	photon2ID = (photonIDVec.size() >= 2) ? photonIDVec[1] : 0;
      }

      if (photon1ID == 0)
      {
	if (m_verboseLevel > 0) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : WARNING : photon1ID == 0" << G4endl;
      }
      if (photon2ID == 0) {
	if (m_verboseLevel > 1) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : WARNING : photon2ID == 0" << G4endl;
      }
      if (m_verboseLevel > 1) G4cout 
	<< "GateAnalysis::RecordEndOfEvent : photon1ID : " << photon1ID 
	<< "     photon2ID : " << photon2ID << G4endl;
    
      
      // analysis of the phantom hits to count the comptons, etc.

      GatePhantomHitsCollection* PHC = GetOutputMgr()->GetPhantomHitCollection();
      NpHits = PHC->entries();
    
      G4String theComptonVolumeName("NULL");
      G4String theComptonVolumeName1("NULL");
      G4String theComptonVolumeName2("NULL");

      G4String theRayleighVolumeName("NULL");
      G4String theRayleighVolumeName1("NULL");
      G4String theRayleighVolumeName2("NULL");

      for (G4int iPHit=0;iPHit<NpHits;iPHit++)
      {
	G4int    phantomTrackID = (*PHC)[iPHit]->GetTrackID();
	G4String processName    = (*PHC)[iPHit]->GetProcess();
	G4int    PDGcode        = (*PHC)[iPHit]->GetPDGEncoding();
	G4ThreeVector hitPos    = (*PHC)[iPHit]->GetPos();

	if (m_verboseLevel > 2) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : GatePhantomHitsCollection : trackID : " << std::setw(5) << phantomTrackID 
	  << "    PDG code : " << std::setw(5) << PDGcode << "  processName : <" << processName << ">" << G4endl;
	theComptonVolumeName = G4String("NULL");

	if (m_verboseLevel > 2) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : GatePhantomHitsCollection : trackID : " << std::setw(5) << phantomTrackID 
	  << "    PDG code : " << std::setw(5) << PDGcode << "  processName : <" << processName << ">" << G4endl;
	theRayleighVolumeName = G4String("NULL");

	// Modif by DS and LS on Oct 4, 2002: we need to be able to recognise both 'compt' 
	// and 'LowEnCompt", hence the find on 'ompt'
//	if (processName.find("ompt") != G4String::npos || processName.find("LowEnRayleigh") != G4String::npos) {
        // modif. by CJG to separate Compton and Rayleigh photons
	if (processName.find("ompt") != G4String::npos)
        {
	  if ((phantomTrackID == photon1ID)||(phantomTrackID == photon2ID))
	  {
	    G4Navigator *gNavigator = G4TransportationManager::GetTransportationManager()->GetNavigatorForTracking();
	    G4ThreeVector null(0.,0.,0.);
	    G4ThreeVector *ptr;
	    ptr = &null;
	    theComptonVolumeName = gNavigator->LocateGlobalPointAndSetup(hitPos,ptr,false)->GetName();
	    if (m_verboseLevel > 1) 
	      G4cout << "GateAnalysis::RecordEndOfEvent :  theComptonVolumeName: " 
		     << theComptonVolumeName << G4endl; 
	  }
	  if (phantomTrackID == photon1ID)
	  {
	    photon1_phantom_compton++;
	    theComptonVolumeName1 = theComptonVolumeName;
	    if (m_verboseLevel > 0) G4cout 
	      << "GateAnalysis::RecordEndOfEvent : photon1_phantom_compton : " << photon1_phantom_compton << G4endl; 
	  }
	  if (phantomTrackID == photon2ID)
	  {
	    photon2_phantom_compton++;
	    theComptonVolumeName2 = theComptonVolumeName;
	    if (m_verboseLevel > 0) G4cout 
	      << "GateAnalysis::RecordEndOfEvent : photon2_phantom_compton : " << photon2_phantom_compton << G4endl; 
	  }
	}
	
	// Counting Rayleigh scatter in phantom
	if (processName.find("LowEnRayleigh") != G4String::npos)
        {
	  if ((phantomTrackID == photon1ID)||(phantomTrackID == photon2ID))
	  {
	    G4Navigator *gNavigator = G4TransportationManager::GetTransportationManager()->GetNavigatorForTracking();
	    G4ThreeVector null(0.,0.,0.);
	    G4ThreeVector *ptr;
	    ptr = &null;
	    theRayleighVolumeName = gNavigator->LocateGlobalPointAndSetup(hitPos,ptr,false)->GetName();
	    if (m_verboseLevel > 1) 
	      G4cout << "GateAnalysis::RecordEndOfEvent :  theRayleighVolumeName: " 
		     << theRayleighVolumeName << G4endl; 
	  }
	  if (phantomTrackID == photon1ID)
	  {
	    photon1_phantom_Rayleigh++;
	    theRayleighVolumeName1 = theRayleighVolumeName;
	    if (m_verboseLevel > 0) G4cout 
	      << "GateAnalysis::RecordEndOfEvent : photon1_phantom_Rayleigh : " << photon1_phantom_Rayleigh << G4endl; 
	  }
	  if (phantomTrackID == photon2ID)
	  {
	    photon2_phantom_Rayleigh++;
	    theRayleighVolumeName2 = theRayleighVolumeName;
	    if (m_verboseLevel > 0) G4cout 
	      << "GateAnalysis::RecordEndOfEvent : photon2_phantom_Rayleigh : " << photon2_phantom_Rayleigh << G4endl; 
	  }
	}

      }
      
  /////////
  // Source info

      G4int sourceID = (((GateSourceMgr::GetInstance())->GetSourcesForThisEvent())[0])->GetSourceID();

      G4ThreeVector sourceVertex = m_trajectoryNavigator->FindSourcePosition();
 
  
  // Hits loop
      for (G4int iHit=0;iHit<NbHits;iHit++)
      {
        
	G4int    crystalTrackID = (*CHC)[iHit]->GetTrackID();
	G4String processName = (*CHC)[iHit]->GetProcess();
        // Counting Compton in the Crystal  
//      if (processName.find("ompt") != G4String::npos || processName.find("LowEnRayleigh") != G4String::npos) {  
        if (processName.find("ompt") != G4String::npos)
        {
	  
	  if (crystalTrackID == photon1ID) photon1_crystal_compton++;
	  if (crystalTrackID == photon2ID) photon2_crystal_compton++;
	   
	}
	
        // Counting Rayleigh scatter in crystal
        if (processName.find("LowEnRayleigh") != G4String::npos)
        {  
	  
	  if (crystalTrackID == photon1ID) photon1_crystal_Rayleigh++;
	  if (crystalTrackID == photon2ID) photon2_crystal_Rayleigh++;
	   
	}
	
	G4int PDGEncoding  = (*CHC)[iHit]->GetPDGEncoding();
	if (m_verboseLevel > 2) G4cout 
	  << "GateAnalysis::RecordEndOfEvent : CrystalHitsCollection: processName : <" << processName 
	  << ">    Particls PDG code : " << PDGEncoding << G4endl;
	if ((*CHC)[iHit]->GoodForAnalysis())
	{
	  // fill in values with the branch with C struct 

	  G4int trackID  = (*CHC)[iHit]->GetTrackID();
	  //G4int parentID = (*CHC)[iHit]->GetParentID();

	  G4int photonID = 0;
	  G4int nPhantomCompton = 0;
	  G4int nCrystalCompton = 0;
	  
	  G4int nPhantomRayleigh = 0;
	  G4int nCrystalRayleigh = 0;	  

	  //	  if ((photon1ID != 0) && (photon2ID != 0)) {
	  if (photon1ID != 0) 
	  { // this means that at least 1 photon has been found, requiring 2 is wrong for SPECT
	    // search the gamma from which this hit comes --> photonID
	    photonID = m_trajectoryNavigator->FindPhotonID(trackID);
	    if (photonID == rootID)
	    {
	      if (m_verboseLevel > 2) G4cout 
		<< "GateAnalysis::RecordEndOfEvent : trackID: " << trackID << " photonID = " << rootID << G4endl; 
	    }
	  }

	  if (photonID == 1)
	  {
	    nPhantomCompton = photon1_phantom_compton;
	    nPhantomRayleigh = photon1_phantom_Rayleigh;
	    theComptonVolumeName = theComptonVolumeName1;
	    theRayleighVolumeName = theRayleighVolumeName1;
	    nCrystalCompton = photon1_crystal_compton;
	    nCrystalRayleigh = photon1_crystal_Rayleigh;
	    
	  }
	  else if (photonID == 2)
	  {
	    nPhantomCompton = photon2_phantom_compton;
	    nPhantomRayleigh = photon2_phantom_Rayleigh;
	    theComptonVolumeName = theComptonVolumeName2;
	    theRayleighVolumeName = theRayleighVolumeName2;
	    nCrystalCompton = photon2_crystal_compton;
	    nCrystalRayleigh = photon2_crystal_Rayleigh;
	    
	  }

	  // search the primary that originated the track
	  primaryID = m_trajectoryNavigator->FindPrimaryID(trackID);
//dk
	  // look for type and energy of primary photon
	  G4int PrimType = m_trajectoryNavigator->FindPrimType(trackID,m_gam1_ID,m_gam2_ID);
	  G4double PrimEnergy = m_trajectoryNavigator->FindPrimEnergy();
G4cout<<" trackID: "<<trackID
      <<"\n primaryID: "<<primaryID
      <<"\n PrimType: "<<PrimType
      <<"\n PrimEnergy: "<<PrimEnergy
      <<"\n m_posID: "<<m_pos_ID  
      <<"\n m_gam1_ID: "<<m_gam1_ID
      <<"\n m_gam2_ID: "<<m_gam2_ID<<G4endl;  
//dk_end	                                

	  (*CHC)[iHit]->SetSourceID          (sourceID);
	  (*CHC)[iHit]->SetSourcePosition    (sourceVertex);
	  (*CHC)[iHit]->SetNPhantomCompton   (nPhantomCompton);
	  (*CHC)[iHit]->SetNPhantomRayleigh   (nPhantomRayleigh);
	  (*CHC)[iHit]->SetComptonVolumeName (theComptonVolumeName);
	  (*CHC)[iHit]->SetRayleighVolumeName (theRayleighVolumeName);
	  (*CHC)[iHit]->SetPhotonID          (photonID);
	  (*CHC)[iHit]->SetPrimaryID         (primaryID);
	  (*CHC)[iHit]->SetEventID           (eventID);
	  (*CHC)[iHit]->SetRunID             (runID);
	  
	  (*CHC)[iHit]->SetNCrystalCompton   (nCrystalCompton); 
	  (*CHC)[iHit]->SetNCrystalRayleigh   (nCrystalRayleigh);
	  

	}
      }

    }
  }
}

//....oooOO0OOooo........oooOO0OOooo........oooOO0OOooo........oooOO0OOooo....

void GateAnalysis::RecordStep(const G4Step* aStep)
{
//dk
   G4VProcess* p=0;
   //note primary positron id
   if(m_pos_ID==0&&aStep!=0){
      G4ParticleDefinition* partDef = aStep->GetTrack()->GetDefinition();
      if (partDef == G4Positron::PositronDefinition()){
         p=(G4VProcess*)aStep->GetTrack()->GetCreatorProcess();
         if(p!=0){
           // R as RadioactiveDecay
           if(p->GetProcessName().index('R')==0) m_pos_ID=aStep->GetTrack()->GetTrackID();
//G4cout<<"<<<3"<<p->GetProcessName()<<" "<<m_pos_ID<<G4endl;
         }
      }
   }
   if(aStep!=0&&aStep->GetTrack()->GetCurrentStepNumber() == 1){
      G4int trackID=aStep->GetTrack()->GetTrackID();
      G4int parentID=aStep->GetTrack()->GetParentID();
      if(trackID!=m_gam1_ID&&trackID!=m_gam2_ID){
         p=(G4VProcess*)aStep->GetTrack()->GetCreatorProcess();
         if(p!=0)if(p->GetProcessName().index('a')==0){
            //we still may have a secondary e+: gamma->e+e- e.g Y86
            if(parentID==m_pos_ID) 
              if(m_gam1_ID==0) {
//G4cout<<"Annih gamma 1 "<<trackID<<G4endl;
                 m_gam1_ID=trackID;
              } else if(m_gam2_ID==0) {
//G4cout<<"Annih gamma 2 "<<trackID<<G4endl;
                 m_gam2_ID=trackID;
              }
         }
      }
   }
//dk_end


  if (m_verboseLevel > 2)
    G4cout << "GateAnalysis::RecordStep" << G4endl;
}

void GateAnalysis::SetVerboseLevel(G4int val) { 
  m_verboseLevel = val; 
  if (m_trajectoryNavigator) m_trajectoryNavigator->SetVerboseLevel(val);
}





